@@ -229,7 +229,7 @@ int gnutls_global_init(void)
 
 	_gnutls_switch_lib_state(LIB_STATE_INIT);
 
-	e = getenv("GNUTLS_DEBUG_LEVEL");
+	e = secure_getenv("GNUTLS_DEBUG_LEVEL");
 	if (e != NULL) {
 		level = atoi(e);
 		gnutls_global_set_log_level(level);
@@ -441,7 +441,7 @@ const char *e;
 	if (_gnutls_global_init_skip() != 0)
 		return;
 
-	e = getenv("GNUTLS_NO_EXPLICIT_INIT");
+	e = secure_getenv("GNUTLS_NO_EXPLICIT_INIT");
 	if (e != NULL) {
 		ret = atoi(e);
 		if (ret == 1)
@@ -462,7 +462,7 @@ static void _DESTRUCTOR lib_deinit(void)
 	if (_gnutls_global_init_skip() != 0)
 		return;
 
-	e = getenv("GNUTLS_NO_EXPLICIT_INIT");
+	e = secure_getenv("GNUTLS_NO_EXPLICIT_INIT");
 	if (e != NULL) {
 		int ret = atoi(e);
 		if (ret == 1)
