@@ -60,6 +60,7 @@ gnutls_pkcs11_copy_x509_crt(const char *token_url,
 	ck_object_class_t class = CKO_CERTIFICATE;
 	ck_certificate_type_t type = CKC_X_509;
 	ck_object_handle_t obj;
+	gnutls_datum_t serial_der = {NULL, 0};
 	int a_val;
 	unsigned long category;
 	struct pkcs11_session_info sinfo;
@@ -152,10 +153,13 @@ gnutls_pkcs11_copy_x509_crt(const char *token_url,
 
 	serial_size = sizeof(serial);
 	if (gnutls_x509_crt_get_serial(crt, serial, &serial_size) >= 0) {
-		a[a_val].type = CKA_SERIAL_NUMBER;
-		a[a_val].value = (void *) serial;
-		a[a_val].value_len = serial_size;
-		a_val++;
+		ret = _gnutls_x509_ext_gen_number(serial, serial_size, &serial_der);
+		if (ret >= 0) {
+			a[a_val].type = CKA_SERIAL_NUMBER;
+			a[a_val].value = (void *) serial_der.data;
+			a[a_val].value_len = serial_der.size;
+			a_val++;
+		}
 	}
 
 	if (label) {
@@ -212,6 +216,7 @@ gnutls_pkcs11_copy_x509_crt(const char *token_url,
 
       cleanup:
 	gnutls_free(der);
+	gnutls_free(serial_der.data);
 	pkcs11_close_session(&sinfo);
 	return ret;
 
@@ -322,11 +327,6 @@ gnutls_pkcs11_copy_x509_privkey(const char *token_url,
 		a_val++;
 	}
 
-	a[a_val].type = CKA_KEY_TYPE;
-	a[a_val].value = &type;
-	a[a_val].value_len = sizeof(type);
-	a_val++;
-
 	a[a_val].type = CKA_TOKEN;
 	a[a_val].value = (void *) &tval;
 	a[a_val].value_len = sizeof(tval);
@@ -470,7 +470,7 @@ gnutls_pkcs11_copy_x509_privkey(const char *token_url,
 			}
 
 			ret =
-			    _gnutls_mpi_dprint_lz(&key->params.
+			    _gnutls_mpi_dprint_lz(key->params.
 						  params[ECC_K], &x);
 			if (ret < 0) {
 				gnutls_assert();
@@ -497,6 +497,11 @@ gnutls_pkcs11_copy_x509_privkey(const char *token_url,
 		goto cleanup;
 	}
 
+	a[a_val].type = CKA_KEY_TYPE;
+	a[a_val].value = &type;
+	a[a_val].value_len = sizeof(type);
+	a_val++;
+
 	rv = pkcs11_create_object(sinfo.module, sinfo.pks, a, a_val, &obj);
 	if (rv != CKR_OK) {
 		gnutls_assert();
