@@ -56,6 +56,8 @@ typedef struct gnutls_x509_crl_int {
 	ASN1_TYPE rcache;
 	unsigned rcache_idx;
 	int use_extensions;
+
+	gnutls_datum_t der;
 	gnutls_datum_t raw_issuer_dn;
 } gnutls_x509_crl_int;
 
@@ -68,7 +70,9 @@ typedef struct gnutls_x509_crt_int {
 	 * get_raw_*_dn(). */
 	gnutls_datum_t raw_dn;
 	gnutls_datum_t raw_issuer_dn;
+	gnutls_datum_t raw_spki;
 
+	gnutls_datum_t der;
 	struct pin_info_st pin;
 } gnutls_x509_crt_int;
 
@@ -81,6 +85,14 @@ typedef struct gnutls_pkcs7_int {
 	unsigned expanded;
 } gnutls_pkcs7_int;
 
+struct pbkdf2_params {
+	uint8_t salt[32];
+	int salt_size;
+	unsigned int iter_count;
+	unsigned int key_size;
+	gnutls_mac_algorithm_t mac;
+};
+
 typedef struct gnutls_x509_privkey_int {
 	/* the size of params depends on the public
 	 * key algorithm
@@ -344,7 +356,8 @@ int _gnutls_pkcs7_decrypt_data(const gnutls_datum_t * data,
 
 typedef enum schema_id {
 	PBES2_GENERIC,		/* when the algorithm is unknown, temporal use when reading only */
-	PBES2_3DES,		/* the stuff in PKCS #5 */
+	PBES2_DES,		/* the stuff in PKCS #5 */
+	PBES2_3DES,
 	PBES2_AES_128,
 	PBES2_AES_192,
 	PBES2_AES_256,
@@ -390,6 +403,7 @@ unsigned int
 _gnutls_pkcs11_verify_crt_status(const char* url,
 				const gnutls_x509_crt_t * certificate_list,
 				unsigned clist_size,
+				const char *purpose,
 				unsigned int flags,
 				gnutls_verify_output_function func);
 #endif
